#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

import sre_yield

# fmt: off
TESTCASES = [
    (5,),
    (1, 5),
    (1, 5, 1),
    (1, 10, 2),
    (10, 1, -1),
    (10, -1, -1),
]
# fmt: on


def test_all():
    for t in TESTCASES:
        yield runner, t


def runner(packed_test):
    expected = list(range(*packed_test))
    print("expected", expected)
    actual = list(sre_yield._bigrange(*packed_test))
    print("actual", actual)
    assert expected == actual


def test_bignum():
    # xrange(start, stop) raises OverflowError in py2.7
    start = sys.maxsize
    stop = sys.maxsize + 5

    l = list(sre_yield._bigrange(start, stop))
    assert len(l) == 5



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import unittest

import sre_yield


class AnchorTest(unittest.TestCase):
    """Test that allowed anchors in allowed positions work."""

    def testAnchorsCaret(self):
        parsed = sre_yield.Values("^[ab]")
        self.assertEqual(["a", "b"], list(parsed))

    def testAnchorsA(self):
        parsed = sre_yield.Values("\\A[ab]")
        self.assertEqual(["a", "b"], list(parsed))

    def testAnchorsMultiCaret(self):
        parsed = sre_yield.Values("^(\\b^([ab]))")
        self.assertEqual(["a", "b"], list(parsed))

    def testParseErrorInMiddle(self):
        self.assertRaises(sre_yield.ParseError, sre_yield.Values, "a\\bb")
        self.assertRaises(sre_yield.ParseError, sre_yield.Values, "a^b")
        self.assertRaises(sre_yield.ParseError, sre_yield.Values, "a$b")

    def testParseErrorNotBoundary(self):
        self.assertRaises(sre_yield.ParseError, sre_yield.Values, "\\Ba")
        self.assertRaises(sre_yield.ParseError, sre_yield.Values, "a\\B")
        self.assertRaises(sre_yield.ParseError, sre_yield.Values, "a\\Bb")

    def testAnchorsDollar(self):
        parsed = sre_yield.Values("[ab]$")
        self.assertEqual(["a", "b"], list(parsed))

    def testAnchorsZ(self):
        parsed = sre_yield.Values("[ab]\\Z")
        self.assertEqual(["a", "b"], list(parsed))

    def testAnchorsCombined(self):
        parsed = sre_yield.Values("^[ab]$")
        self.assertEqual(["a", "b"], list(parsed))

    def testAnchorsBoundary(self):
        parsed = sre_yield.Values("ab\\b")
        self.assertEqual(["ab"], list(parsed))
        parsed = sre_yield.Values("\\bab")
        self.assertEqual(["ab"], list(parsed))
        parsed = sre_yield.Values("\\bab\\b")
        self.assertEqual(["ab"], list(parsed))

    def testAnchorsRepeated(self):
        parsed = sre_yield.Values(r"^\b^[ab]")
        self.assertEqual(["a", "b"], list(parsed))

    def testAnchorsRepeated(self):
        parsed = sre_yield.Values(r"[ab]$\b$")
        self.assertEqual(["a", "b"], list(parsed))

    def testAnchorsEmptyString(self):
        parsed = sre_yield.Values(r"^^^$\b$")
        self.assertEqual([""], list(parsed))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
# Copyright 2018-2019 Tim Hatch
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import io
import re
import sys
import unittest

import sre_yield


class YieldTest(unittest.TestCase):
    """Test that regular expressions give the right lists."""

    def testSimpleCases(self):
        self.assertSequenceEqual(
            sre_yield.AllStrings("1(234?|49?)"), ["123", "1234", "14", "149"]
        )
        self.assertSequenceEqual(sre_yield.AllStrings("asd|def"), ["asd", "def"])
        self.assertSequenceEqual(
            sre_yield.AllStrings("asd|def\\+|a\\.b\\.c"), ["asd", "def+", "a.b.c"]
        )

    def testOtherCases(self):
        self.assertSequenceEqual(sre_yield.AllStrings("[aeiou]"), list("aeiou"))
        self.assertEqual(len(sre_yield.AllStrings("1.3", flags=re.DOTALL)), 256)
        v = sre_yield.AllStrings("[^-]3[._]1415", flags=re.DOTALL)
        print(list(v))
        self.assertEqual(len(v), 510)
        self.assertEqual(
            len(sre_yield.AllStrings("(.|5[6-9]|[6-9][0-9])[a-z].?", flags=re.DOTALL)),
            300 * 26 * 257,
        )
        self.assertEqual(len(sre_yield.AllStrings("..", charset="0123456789")), 100)
        self.assertEqual(len(sre_yield.AllStrings("0*")), 65536)
        # For really big lists, we can't use the len() function any more
        self.assertEqual(sre_yield.AllStrings("0*").__len__(), 65536)
        self.assertEqual(sre_yield.AllStrings("[01]*").__len__(), 2 ** 65536 - 1)

    def testAlternationWithEmptyElement(self):
        self.assertSequenceEqual(sre_yield.AllStrings("a(b|c|)"), ["ab", "ac", "a"])
        self.assertSequenceEqual(sre_yield.AllStrings("a(|b|c)"), ["a", "ab", "ac"])
        self.assertSequenceEqual(sre_yield.AllStrings("a[bc]?"), ["a", "ab", "ac"])
        self.assertSequenceEqual(sre_yield.AllStrings("a[bc]??"), ["a", "ab", "ac"])

    def testSlices(self):
        parsed = sre_yield.AllStrings("[abcdef]")
        self.assertSequenceEqual(parsed[::2], list("ace"))
        self.assertSequenceEqual(parsed[1::2], list("bdf"))
        self.assertSequenceEqual(parsed[1:-1], list("bcde"))
        self.assertSequenceEqual(parsed[1:-2], list("bcd"))
        self.assertSequenceEqual(parsed[1:99], list("bcdef"))
        self.assertSequenceEqual(parsed[1:1], [])

        self.assertEqual(parsed[1], "b")
        self.assertEqual(parsed[-2], "e")
        self.assertEqual(parsed[-1], "f")

    def testSlicesRepeated(self):
        parsed = sre_yield.AllStrings("[abcdef]")
        self.assertSequenceEqual(parsed[::-1][:2], list("fe"))
        self.assertSequenceEqual(parsed[1:][1:][1:-1], list("de"))
        self.assertSequenceEqual(parsed[::2][1:], list("ce"))

    def testGetItemNegative(self):
        parsed = sre_yield.AllStrings("x|[a-z]{1,5}")
        self.assertEqual(parsed[0], "x")
        self.assertEqual(parsed[1], "a")
        self.assertEqual(parsed[23], "w")
        self.assertEqual(parsed[24], "x")
        self.assertEqual(parsed[25], "y")
        self.assertEqual(parsed[26], "z")
        self.assertEqual(parsed[27], "aa")
        self.assertEqual(parsed[28], "ab")
        self.assertEqual(parsed[-2], "zzzzy")
        self.assertEqual(parsed[-1], "zzzzz")

        # last, and first
        parsed.get_item(len(parsed) - 1)
        parsed.get_item(-len(parsed))

        # precisely 1 out of bounds
        self.assertRaises(IndexError, parsed.get_item, len(parsed))
        self.assertRaises(IndexError, parsed.get_item, -len(parsed) - 1)

    def testContains(self):
        parsed = sre_yield.AllStrings("[01]+")
        self.assertTrue("0101" in parsed)
        self.assertFalse("0201" in parsed)

    def testNaturalOrder(self):
        parsed = sre_yield.AllStrings("[0-9]{2}")
        self.assertEqual(parsed[0], "00")
        self.assertEqual(parsed[1], "01")
        self.assertEqual(parsed[98], "98")
        self.assertEqual(parsed[99], "99")

    def testCategories(self):
        cat_chars = "wWdDsS"
        for c in cat_chars:
            r = re.compile("\\" + c, flags=sre_yield.DEFAULT_RE_FLAGS)
            matching = [i for i in sre_yield.CHARSET if r.match(i)]
            self.assertGreater(len(matching), 5)
            parsed = sre_yield.AllStrings("\\" + c)
            self.assertEqual(sorted(matching), sorted(parsed[:]))

    def testDotallFlag(self):
        parsed = sre_yield.AllStrings(".", charset="abc\n")
        self.assertEqual(["a", "b", "c"], parsed[:])
        parsed = sre_yield.AllStrings(".", charset="abc\n", flags=re.DOTALL)
        self.assertEqual(["a", "b", "c", "\n"], parsed[:])

    def testMaxCount(self):
        parsed = sre_yield.AllStrings("[01]+", max_count=4)
        self.assertEqual("1111", parsed[-1])

    def testParseErrors(self):
        self.assertRaises(sre_yield.ParseError, sre_yield.AllStrings, "a", re.I)
        self.assertRaises(sre_yield.ParseError, sre_yield.AllStrings, "a", re.U)
        # Causes a failure inside sre_parse under Python 3.6
        # self.assertRaises(sre_yield.ParseError, sre_yield.AllStrings, 'a', re.L)

    def testSavingGroups(self):
        parsed = sre_yield.AllStrings(r"(([abc])d)e")
        d = {}
        self.assertEqual("ade", parsed.get_item(0, d))
        self.assertEqual("ad", d[1])
        self.assertEqual("a", d[2])

    def testSavingGroupsByName(self):
        parsed = sre_yield.AllMatches(r"x(?P<foo>[abc])x")
        m = parsed[0]
        self.assertEqual("xax", m.group(0))
        self.assertEqual("a", m.group(1))
        self.assertEqual("a", m.group("foo"))
        self.assertEqual({"foo": "a"}, m.groupdict())
        self.assertRaises(NotImplementedError, m.span)

    def testBackrefCounts(self):
        parsed = sre_yield.AllStrings(r"([abc])-\1")
        self.assertEqual(3, len(parsed))
        self.assertEqual(["a-a", "b-b", "c-c"], parsed[:])

    def testSlicingMatches(self):
        parsed = sre_yield.AllMatches(r"([abc])-\1")
        self.assertEqual(3, len(parsed))
        self.assertEqual(["a-a", "b-b"], [x.group(0) for x in parsed[:2]])

    def testAllStringsIsValues(self):
        self.assertEqual(sre_yield.AllStrings, sre_yield.Values)

    def testCanIterateGiantValues(self):
        v = sre_yield.AllStrings(".+")
        self.assertGreater(v.__len__(), sys.maxsize)
        it = iter(v)
        self.assertEqual("\x00", next(it))
        self.assertEqual("\x01", next(it))

    def testCanSliceGiantValues(self):
        v = sre_yield.AllStrings(".+")
        self.assertGreater(v.__len__(), sys.maxsize)
        self.assertEqual(["\x00", "\x01"], list(v[:2]))

    def testOffset(self):
        # This was discovered after https://github.com/google/sre_yield/issues/10
        v = sre_yield.AllStrings("([0-9a-fA-F]{0,4}:){0,5}")
        l = v.__len__()
        self.assertTrue(v.__getitem__(l - 1))

    def testMain(self):
        old_sys_stdout = sys.stdout

        # This is for compatibility with python 2.7
        buf = io.BytesIO()
        try:
            sys.stdout = codecs.lookup("utf-8").streamwriter(buf)
            sre_yield.main(["prog", "x[123]"])
            value = sys.stdout.getvalue()
        finally:
            sys.stdout = old_sys_stdout

        self.assertEqual(b"x1\nx2\nx3\n", buf.getvalue())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from sre_yield import cachingseq


class CachingFuncSequenceTest(unittest.TestCase):
    def testLimits(self):
        c = cachingseq.CachingFuncSequence(lambda i: i, 10)
        self.assertEqual(9, c[9])
        self.assertEqual(9, c[-1])
        self.assertEqual(0, c[0])
        self.assertEqual(0, c[-10])
        self.assertRaises(IndexError, lambda: c[10])
        self.assertRaises(IndexError, lambda: c[11])
        self.assertRaises(IndexError, lambda: c[-11])
        self.assertRaises(IndexError, lambda: c[-12])
        self.assertEqual(2, len(c._cache))

        # Make sure .func is settable at runtime...
        c.func = lambda i: "bbb"
        self.assertEqual("bbb", c[1])
        # ...and that we don't call it again.
        self.assertEqual(0, c[0])

    def testIter(self):
        c = cachingseq.CachingFuncSequence(lambda i: i, 10)
        # Cache empty on construction
        self.assertEqual(0, len(c._cache))

        self.assertEqual(10, len(c))
        self.assertEqual(list(range(10)), list(c))

        # Cache full after iteration
        self.assertEqual(10, len(c._cache))

    def testIncFunc(self):
        def first_func(x):
            assert x == 0
            return 1

        def inc_func(i, prev):
            return prev * 2

        c = cachingseq.CachingFuncSequence(first_func, 10, inc_func)
        self.assertEqual([1, 2, 4, 8, 16, 32, 64, 128, 256, 512], list(c))



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
# Copyright 2018 Tim Hatch
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# vim: sw=2 sts=2 et

"""This module can generate all strings that match a regular expression.

The regex is parsed using the SRE module that is standard in python,
then the data structure is executed to form a bunch of iterators.
"""

__author__ = "alexperry@google.com (Alex Perry)"
__all__ = ["Values", "AllStrings", "AllMatches", "ParseError"]


import bisect
import math
import re
import sre_constants
import sre_parse
import string
import sys
import types

from sre_yield import cachingseq, fastdivmod

try:
    xrange = xrange
except NameError:
    xrange = range

_RE_METACHARS = r"$^{}*+\\"
_ESCAPED_METACHAR = r"\\[" + _RE_METACHARS + r"]"
ESCAPED_METACHAR_RE = re.compile(_ESCAPED_METACHAR)
# ASCII by default, see https://github.com/google/sre_yield/issues/3
CHARSET = [chr(c) for c in range(256)]

WORD = string.ascii_letters + string.digits + "_"

try:
    DEFAULT_RE_FLAGS = re.ASCII
except AttributeError:
    DEFAULT_RE_FLAGS = 0

STATE_START, STATE_MIDDLE, STATE_END = list(range(3))


def Not(chars):
    return "".join(sorted(set(CHARSET) - set(chars)))


CATEGORIES = {
    sre_constants.CATEGORY_WORD: WORD,
    sre_constants.CATEGORY_NOT_WORD: Not(WORD),
    sre_constants.CATEGORY_DIGIT: string.digits,
    sre_constants.CATEGORY_NOT_DIGIT: Not(string.digits),
    sre_constants.CATEGORY_SPACE: string.whitespace,
    sre_constants.CATEGORY_NOT_SPACE: Not(string.whitespace),
}

# This constant varies between builds of Python; this is the lower value.
MAX_REPEAT_COUNT = 65535


class ParseError(Exception):
    pass


def slice_indices(slice_obj, size):
    """slice_obj.indices() except this one supports longs."""
    # start stop step
    start = slice_obj.start
    stop = slice_obj.stop
    step = slice_obj.step

    # We don't always update a value for negative indices (if we wrote it here
    # due to None).
    if step is None:
        step = 1
    if start is None:
        if step > 0:
            start = 0
        else:
            start = size - 1
    else:
        start = _adjust_index(start, size)

    if stop is None:
        if step > 0:
            stop = size
        else:
            stop = -1
    else:
        stop = _adjust_index(stop, size)

    return (start, stop, step)


def _adjust_index(n, size):
    if n < 0:
        n += size

    if n < 0:
        raise IndexError("Out of range")
    if n > size:
        n = size
    return n


def _xrange(*args):
    """Because xrange doesn't support longs :("""
    # prefer real xrange if it works
    try:
        return xrange(*args)
    except OverflowError:
        return _bigrange(*args)


def _bigrange(*args):
    if len(args) == 1:
        start = 0
        stop = args[0]
        step = 1
    elif len(args) == 2:
        start, stop = args
        step = 1
    elif len(args) == 3:
        start, stop, step = args
    else:
        raise ValueError("Too many args for _bigrange")

    i = start
    while True:
        yield i
        i += step
        if step < 0 and i <= stop:
            break
        if step > 0 and i >= stop:
            break


class WrappedSequence(object):
    """This wraps a sequence, purely as a base clase for the other uses."""

    def __init__(self, raw):
        # Derived classes will likely override this constructor
        self.raw = raw
        # Note that we can't use the function len() because it insists on trying
        # to convert the returned number from a long-int to an ordinary int.
        self.length = raw.__len__()

    def get_item(self, i, d=None):
        i = _adjust_index(i, self.length)
        if hasattr(self.raw, "get_item"):
            return self.raw.get_item(i, d)
        return self.raw[i]

    def __len__(self):
        return self.length

    def __getitem__(self, i):
        # If the user wanted a slice, we provide a wrapper
        if isinstance(i, slice):
            result = SlicedSequence(self, slicer=i)
            if result.__len__() < 16:
                # Short lists are unpacked
                result = [item for item in result]
            return result
        i = _adjust_index(i, self.length)
        # Usually we just call the user-provided function
        return self.get_item(i)

    def __iter__(self):
        for i in _xrange(int(self.length)):
            yield self.get_item(i)


def _sign(x):
    if x > 0:
        return 1
    else:
        return -1


class SlicedSequence(WrappedSequence):
    """This is part of an immutable and potentially arbitrarily long list."""

    def __init__(self, raw, slicer=None):
        # Derived classes will likely override this constructor
        self.raw = raw
        if slicer is None:
            self.start, self.stop, self.step = 0, raw.__len__(), 1
        else:
            self.start, self.stop, self.step = slice_indices(slicer, raw.__len__())

        # Integer round up, depending on step direction
        self.length = (
            self.stop - self.start + self.step - _sign(self.step)
        ) / self.step

    def get_item(self, i, d=None):
        j = i * self.step + self.start
        return self.raw[j]


class ConcatenatedSequence(WrappedSequence):
    """This is equivalent to using extend() but without unpacking the lists."""

    def __init__(self, *alternatives):
        self.list_lengths = [(a, a.__len__()) for a in alternatives]
        self.length = sum(a_len for _, a_len in self.list_lengths)

    def get_item(self, i, d=None):
        for a, a_len in self.list_lengths:
            if i < a_len:
                return a[i]
            i -= a_len
        raise IndexError("Too Big")

    def __contains__(self, item):
        for a, _ in self.list_lengths:
            if item in a:
                return True
        return False

    def __repr__(self):
        return "{concat " + repr(self.list_lengths) + "}"


class CombinatoricsSequence(WrappedSequence):
    """This uses all combinations of one item from each passed list."""

    def __init__(self, *components):
        self.list_lengths = [(a, a.__len__()) for a in components]
        self.length = 1
        for _, c_len in self.list_lengths:
            self.length *= c_len

    def get_item(self, i, d=None):
        result = []
        if i < 0:
            i += self.length
        if i < 0 or i >= self.length:
            raise IndexError("Index %d out of bounds" % (i,))

        if len(self.list_lengths) == 1:
            # skip unnecessary ''.join -- big speedup
            return self.list_lengths[0][0][i]

        for c, c_len in self.list_lengths:
            i, mod = divmod(i, c_len)
            if hasattr(c, "get_item"):
                result.append(c.get_item(mod, d))
            else:
                result.append(c[mod])
        return "".join(result)

    def __repr__(self):
        return "{combin " + repr(self.list_lengths) + "}"


# Intuition is that this should be around 2**16 to 2**64, but the exact value is
# not important.
OFFSET_BREAK_THRESHOLD = sys.maxsize


class RepetitiveSequence(WrappedSequence):
    """This chooses an entry from a list, many times, and concatenates."""

    def __init__(self, content, lowest=1, highest=1):
        self.content = content
        self.content_length = content.__len__()
        self.length = fastdivmod.powersum(self.content_length, lowest, highest)
        self.lowest = lowest
        self.highest = highest

        def arbitrary_entry(i):
            return (
                fastdivmod.powersum(self.content_length, lowest, i + lowest - 1),
                i + lowest,
            )

        def entry_from_prev(i, prev):
            return (prev[0] + (self.content_length ** prev[1]), prev[1] + 1)

        self.offsets = cachingseq.CachingFuncSequence(
            arbitrary_entry, highest - lowest + 1, entry_from_prev
        )

        # `offset_break` is an optimization around bisect, which would normally
        # choose the "middle" value to bisect on, which does a lot of work
        # that's unnecessary at the bottom of the range (say, the first 256
        # entries).
        #
        # A good choice of OFFSET_BREAK_THRESHOLD minimizes the wasted work up
        # front (we have to calculate all the offsets now up to it), and is
        # still larger than most performant lookups will need.  Anything above
        # that will result in a big penalty as we get into arbitrary-precision
        # integers and use the standard bisect logic.

        if self.offsets[-1][0] > sys.maxsize:
            for i in range(len(self.offsets) - 1):
                if self.offsets[i + 1][0] > sys.maxsize:
                    self.index_of_offset = i
                    self.offset_break = self.offsets[i][0]
                    return

        self.index_of_offset = len(self.offsets)
        self.offset_break = self.offsets[-1][0] + 1

    def get_item(self, i, d=None):
        """Finds out how many repeats this index implies, then picks strings."""
        if i < self.offset_break:
            by_bisect = bisect.bisect_left(
                self.offsets, (i, -1), hi=self.index_of_offset
            )
        else:
            by_bisect = bisect.bisect_left(
                self.offsets, (i, -1), lo=self.index_of_offset
            )

        if by_bisect == len(self.offsets) or self.offsets[by_bisect][0] > i:
            by_bisect -= 1

        num = i - self.offsets[by_bisect][0]
        count = self.offsets[by_bisect][1]

        if count > 100 and self.content_length < 1000:
            content = list(self.content)
        else:
            content = self.content

        result = []

        if count == 0:
            return ""

        for modulus in fastdivmod.divmod_iter(num, self.content_length):
            result.append(content[modulus])

        leftover = count - len(result)
        if leftover:
            assert leftover > 0
            result.extend([content[0]] * leftover)

        # smallest place value ends up on the right
        return "".join(result[::-1])

    def __repr__(self):
        return "{repeat base=%d low=%d high=%d}" % (
            self.content_length,
            self.lowest,
            self.highest,
        )


class SaveCaptureGroup(WrappedSequence):
    def __init__(self, parsed, key):
        self.key = key
        super(SaveCaptureGroup, self).__init__(parsed)

    def get_item(self, n, d=None):
        rv = super(SaveCaptureGroup, self).get_item(n, d)
        if d is not None:
            d[self.key] = rv
        return rv


class ReadCaptureGroup(WrappedSequence):
    def __init__(self, n):
        self.num = n
        self.length = 1

    def get_item(self, i, d=None):
        if i != 0:
            raise IndexError(i)
        if d is None:
            raise ValueError("ReadCaptureGroup with no dict")
        return d.get(self.num, "fail")


class RegexMembershipSequence(WrappedSequence):
    """Creates a sequence from the regex, knows how to test membership."""

    def empty_list(self, *_):
        return []

    def nothing_added(self, *_):
        return [""]

    def branch_values(self, _, items):
        """Converts SRE parser data into literals and merges those lists."""
        return ConcatenatedSequence(*[self.sub_values(parsed) for parsed in items])

    def max_repeat_values(self, min_count, max_count, items):
        """Sequential expansion of the count to be combinatorics."""
        max_count = min(max_count, self.max_count)
        return RepetitiveSequence(self.sub_values(items), min_count, max_count)

    def in_values(self, items):
        # Special case which distinguishes branch from charset operator
        if items and items[0][0] == sre_constants.NEGATE:
            items = self.branch_values(None, items[1:])
            return [item for item in self.charset if item not in items]
        return self.branch_values(None, items)

    def not_literal(self, y):
        return self.in_values(((sre_constants.NEGATE,), (sre_constants.LITERAL, y)))

    def category(self, y):
        return CATEGORIES[y]

    def groupref(self, n):
        self.has_groupref = True
        return ReadCaptureGroup(n)

    def get_item(self, i, d=None):
        """Typically only pass i.  d is an internal detail, for consistency with other classes.

        If you care about the capture groups, you should use
        RegexMembershipSequenceMatches instead, which returns a Match object
        instead of a string."""
        if self.has_groupref or d is not None:
            if d is None:
                d = {}
            return super(RegexMembershipSequence, self).get_item(i, d)
        else:
            return super(RegexMembershipSequence, self).get_item(i)

    def sub_values(self, parsed):
        """This knows how to convert one piece of parsed pattern."""
        # If this is a subpattern object, we just want its data
        if isinstance(parsed, sre_parse.SubPattern):
            parsed = parsed.data
        # A list indicates sequential elements of a string
        if isinstance(parsed, list):
            elements = [self.sub_values(p) for p in parsed]
            return CombinatoricsSequence(*elements)
        # If not a list, a tuple represents a specific match type
        if isinstance(parsed, tuple) and parsed:
            matcher, arguments = parsed
            if not isinstance(arguments, tuple):
                arguments = (arguments,)
            if matcher in self.backends:
                self.check_anchor_state(matcher, arguments)
                return self.backends[matcher](*arguments)
        # No idea what to do here
        raise ParseError(repr(parsed))

    def maybe_save(self, *args):
        # Python 3.6 has group, add_flags, del_flags, parsed
        # while earlier versions just have group, parsed
        group = args[0]
        parsed = args[-1]
        rv = self.sub_values(parsed)
        if group is not None:
            rv = SaveCaptureGroup(rv, group)
        return rv

    def check_anchor_state(self, matcher, arguments):
        # A bit of a hack to support zero-width leading anchors.  The goal is
        # that /^(a|b)$/ will match properly, and that /a^b/ or /a\bb/ throws
        # an error.  (It's unfortunate that I couldn't easily handle /$^/ which
        # matches the empty string; I went for the common case.)
        #
        # There are three states, for example:
        # / STATE_START
        # | / STATE_START (^ causes no transition here, but is illegal at STATE_MIDDLE or STATE_END)
        # | |  / STATE_START (\b causes no transition here, but advances MIDDLE to END)
        # | |  | / (same as above for ^)
        # | |  | | / STATE_MIDDLE (anything besides ^ and \b advances START to MIDDLE)
        # | |  | | | / still STATE_MIDDLE
        # . .  . . . .  / advances MIDDLE to END
        #  ^ \b ^ X Y \b $
        if self.state == STATE_START:
            if matcher == sre_constants.AT:
                if arguments[0] in (sre_constants.AT_END, sre_constants.AT_END_STRING):
                    self.state = STATE_END
                elif arguments[0] == sre_constants.AT_NON_BOUNDARY:
                    # This is nonsensical at beginning of string
                    raise ParseError("Anchor %r found at START state" % (arguments[0],))
                # All others (AT_BEGINNING, AT_BEGINNING_STRING, and AT_BOUNDARY) remain in START.
            elif matcher != sre_constants.SUBPATTERN:
                self.state = STATE_MIDDLE
            # subpattern remains in START
        elif self.state == STATE_END:
            if matcher == sre_constants.AT:
                if arguments[0] not in (
                    sre_constants.AT_END,
                    sre_constants.AT_END_STRING,
                    sre_constants.AT_BOUNDARY,
                ):
                    raise ParseError("Anchor %r found at END state" % (arguments[0],))
                # those three remain in END
            elif matcher != sre_constants.SUBPATTERN:
                raise ParseError(
                    "Non-end-anchor %r found at END state" % (arguments[0],)
                )
            # subpattern remains in END
        else:  # self.state == STATE_MIDDLE
            if matcher == sre_constants.AT:
                if arguments[0] not in (
                    sre_constants.AT_END,
                    sre_constants.AT_END_STRING,
                    sre_constants.AT_BOUNDARY,
                ):
                    raise ParseError(
                        "Anchor %r found at MIDDLE state" % (arguments[0],)
                    )
                # All others (AT_END, AT_END_STRING, AT_BOUNDARY) advance to END.
                self.state = STATE_END

    def __init__(self, pattern, flags=0, charset=CHARSET, max_count=None):
        # If the RE module cannot compile it, we give up quickly
        self.matcher = re.compile(r"(?:%s)\Z" % pattern, flags)
        if not flags & re.DOTALL:
            charset = "".join(c for c in charset if c != "\n")
        self.charset = charset

        self.named_group_lookup = self.matcher.groupindex

        flags |= DEFAULT_RE_FLAGS  # https://github.com/google/sre_yield/issues/3
        if flags & re.IGNORECASE:
            raise ParseError(
                'Flag "i" not supported. https://github.com/google/sre_yield/issues/4'
            )
        elif flags & re.UNICODE:
            raise ParseError(
                'Flag "u" not supported. https://github.com/google/sre_yield/issues/3'
            )
        elif flags & re.LOCALE:
            raise ParseError(
                'Flag "l" not supported. https://github.com/google/sre_yield/issues/5'
            )

        if max_count is None:
            self.max_count = MAX_REPEAT_COUNT
        else:
            self.max_count = max_count

        self.has_groupref = False

        # Configure the parser backends
        self.backends = {
            sre_constants.LITERAL: lambda y: [chr(y)],
            sre_constants.RANGE: lambda l, h: [chr(c) for c in range(l, h + 1)],
            sre_constants.SUBPATTERN: self.maybe_save,
            sre_constants.BRANCH: self.branch_values,
            sre_constants.MIN_REPEAT: self.max_repeat_values,
            sre_constants.MAX_REPEAT: self.max_repeat_values,
            sre_constants.AT: self.nothing_added,
            sre_constants.ASSERT: self.empty_list,
            sre_constants.ASSERT_NOT: self.empty_list,
            sre_constants.ANY: lambda _: self.in_values(((sre_constants.NEGATE,),)),
            sre_constants.IN: self.in_values,
            sre_constants.NOT_LITERAL: self.not_literal,
            sre_constants.CATEGORY: self.category,
            sre_constants.GROUPREF: self.groupref,
        }
        self.state = STATE_START
        # Now build a generator that knows all possible patterns
        self.raw = self.sub_values(sre_parse.parse(pattern, flags))
        # Configure this class instance to know about that result
        self.length = self.raw.__len__()

    def __contains__(self, item):
        # Since we have a regex, we can search the list really cheaply
        return self.matcher.match(item) is not None


class RegexMembershipSequenceMatches(RegexMembershipSequence):
    def __getitem__(self, i):
        if isinstance(i, slice):
            result = SlicedSequence(self, slicer=i)
            if result.__len__() < 16:
                # Short lists are unpacked
                result = [item for item in result]
            return result

        d = {}
        s = super(RegexMembershipSequenceMatches, self).get_item(i, d)
        return Match(s, d, self.named_group_lookup)


def AllStrings(regex, flags=0, charset=CHARSET, max_count=None):
    """Constructs an object that will generate all matching strings."""
    return RegexMembershipSequence(regex, flags, charset, max_count=max_count)


Values = AllStrings


class Match(object):
    def __init__(self, string, groups, named_groups):
        # TODO keep group(0) only, and spans for the rest.
        self._string = string
        self._groups = groups
        self._named_groups = named_groups
        self.lastindex = len(groups) + 1

    def group(self, n=0):
        if n == 0:
            return self._string
        if not isinstance(n, int):
            n = self._named_groups[n]
        return self._groups[n]

    def groups(self):
        return tuple(self._groups[i] for i in range(1, self.lastindex))

    def groupdict(self):
        d = {}
        for k, v in self._named_groups.items():
            d[k] = self._groups[v]
        return d

    def span(self, n=0):
        raise NotImplementedError()


def AllMatches(regex, flags=0, charset=CHARSET, max_count=None):
    """Constructs an object that will generate all matching strings."""
    return RegexMembershipSequenceMatches(regex, flags, charset, max_count=max_count)


def main(argv=None):
    """This module can be executed on the command line for testing."""
    if argv is None:
        argv = sys.argv
    for arg in argv[1:]:
        for i in AllStrings(arg):
            print(i)


if __name__ == "__main__":
    main()



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import unittest

import sre_yield


class ExtractSliceType(object):
    """
    This exists so we can index obj[a:b:c] and get back a slice type without
    worrying about whether its arguments get mangled.
    """

    def __getitem__(self, n):
        return n


E = ExtractSliceType()

# Confirm that the new slice_indices function behaves like SliceType.indices

ARR = list(map(str, list(range(100))))
REG = sre_yield.AllStrings("\d{1}|1\d|2\d|3\d|4\d|5\d|6\d|7\d|8\d|9\d")


def test_prereqs():
    # TODO: Order of sre_yield may change at some point, to increment LSB
    # first.
    assert ARR == list(REG)


NUMS = [None, 0, 2, 5, 80, 90, -20, -10, -1, 100, 110]

TESTCASES = list(itertools.combinations(NUMS, 3))


def test_parity():
    for (start, stop, step) in TESTCASES:
        yield indices_runner, start, stop, step
        yield content_runner, start, stop, step


def indices_runner(start, stop, step):
    st = E[start:stop:step]
    expected = st.indices(100)
    print("expected", expected)
    actual = sre_yield.slice_indices(st, 100)
    print("actual", actual)
    print("array", ARR[st])
    assert expected == actual


def content_runner(start, stop, step):
    st = E[start:stop:step]
    indices = st.indices(100)
    print("indices", indices)
    expected = ARR[st]
    print("expected", expected)
    actual = list(REG[st])
    print("actual", actual)
    assert expected == actual



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# vim: sw=2 sts=2 et


class CachingFuncSequence(object):
    def __init__(self, func, length, inc_func=None):
        """
        length: Length of this sequence.
        func: function(index)
        inc_func: function(index, value_of_previous)
        """

        self.func = func
        self.inc_func = inc_func
        self.length = length
        self._cache = {}

    def __getitem__(self, i):
        if i < 0:
            i += self.length
        if i < 0 or i >= self.length:
            raise IndexError()

        v = self._cache.get(i)
        if v is not None:
            return v

        if self.inc_func and i - 1 in self._cache:
            v = self.inc_func(i, self._cache[i - 1])
        else:
            v = self.func(i)

        self._cache[i] = v
        return v

    def __len__(self):
        return self.length

    def __iter__(self):
        for i in range(self.length):
            yield self[i]



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
# Copyright 2018 Tim Hatch
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# vim: sw=2 sts=2 et

import os.path

from setuptools import find_packages, setup

THIS_DIR = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(THIS_DIR, "README.rst")).read()

VERSION = "1.2"

setup(
    name="sre_yield",
    version=VERSION,
    description="Expands a regular expression to its possible matches",
    long_description=README,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
    ],
    keywords="",
    author="Alex Perry",
    author_email="alex.perry@google.com",
    url="https://github.com/google/sre_yield",
    license="Apache",
    packages=find_packages("."),
    install_requires=[],
    entry_points={"console_scripts": ["demo_sre_yield=sre_yield:main"]},
    test_suite="nose.collector",  # doesn't find doctests though
    tests_require=["nose"],
)



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import unittest

import sre_yield


class SlowYieldTest(unittest.TestCase):
    """Test that regular expressions give the right lists."""

    def testDotStarCase(self):
        test_size = sre_yield.AllStrings(".*", re.DOTALL).__len__()
        actual_size = 0
        for _ in range(65536):
            actual_size = actual_size * 256 + 1
        self.assertEqual(test_size, actual_size)

    def testContentsNearBreak(self):
        # This specific location is on either side of offset_break in sre_yield.py
        v = sre_yield.Values(".*")
        a = v[70386233652806655][:10]
        self.assertEqual("\xff\xff\xff\xff\xff\xff\xff", a)
        b = v[70386233652806656][:10]
        self.assertEqual("\x00\x00\x00\x00\x00\x00\x00\x00", b)
        c = v[70386233652806657][:10]
        self.assertEqual("\x00\x00\x00\x00\x00\x00\x00\x01", c)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python2
#
# Copyright 2011-2016 Google Inc.
# Copyright 2018-2019 Tim Hatch
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# vim: sw=2 sts=2 et

from math import log

try:
    long = long  # py2.7 compat
except NameError:
    long = int

__all__ = ["divmod_iter", "find_largest_power", "powersum"]


def find_largest_power(less_than, base):
    """
    Returns the largest power of `base` that is less than or equal to
    `less_than`.
    """
    # avoid div by zero
    if less_than == 0:
        return 0
    power = int(log(less_than) / log(base))
    return base ** power


def divmod_iter(x, by, chunk=None):
    """
    Generate successive (x % by); x /= by
    """
    if not isinstance(x, (int, long)):
        raise TypeError("`x` must be an int")
    if not isinstance(by, (int, long)):
        raise TypeError("`by` must be an int")
    if chunk is not None and not isinstance(chunk, (int, long)):
        raise TypeError("`chunk` must be an int")

    if x < by:
        return [x]

    if hasattr(x, "bit_length"):
        # crude log(2, x)
        divisions = x.bit_length() // by.bit_length()
    else:
        # This code path is intended for ints, but on <2.7 or alternate
        # implementation that does not support bit_length.  It is not covered by
        # current tests.
        divisions = log(x) // log(by)

    if divisions < 1024:
        return divmod_iter_basic(x, by, chunk)
    else:
        return divmod_iter_chunking(x, by, chunk)


def divmod_iter_chunking(x, by, chunk=None):
    """Generate successive (x % by); x /= by, but faster.

    If provided, |chunk| must be a power of |by| (otherwise it is determined
    automatically for 1024 per inner loop, based on analysis of bench_fastdivmod.py)
    """

    if by == 1:
        if x != 0:
            raise ValueError(
                "x=0 by=1 is allowed as a base case, but no other x may have " "by=1"
            )
        yield 0
        return

    if chunk is None:
        digits_per_chunk = 1024
        chunk = by ** digits_per_chunk
    else:
        digits_per_chunk = int(round(log(chunk) / log(by)))
        if (by ** digits_per_chunk) != chunk:
            raise ValueError("Chunk=%d must be a power of by=%d" % (chunk, by))

    assert digits_per_chunk > 0

    while x:
        x, this_chunk = divmod(x, chunk)
        # this_chunk = int(this_chunk)
        for _ in range(digits_per_chunk):
            this_chunk, m = divmod(this_chunk, by)
            yield m

            if this_chunk == 0 and x == 0:
                break


def divmod_iter_basic(x, by, chunk=None):
    """Generate successive (x % by); x /= by, the obvious way.

    Chunk is ignored.
    """
    while x:
        x, m = divmod(x, by)
        yield m


def powersum(x, low, high):
    # http://mikestoolbox.com/powersum.html
    xm1 = x - 1
    if xm1 == 0:
        return high - low + 1
    a = x ** (high + 1)
    b = x ** low
    return (a - b) // xm1
